@@ -1,22 +1,20 @@
 # -*- coding: future_fstrings -*-
 
-import colorsys
 import logging
 import time
 
 import pigpio
-import smbus2
+from smbus2 import SMBus
 from PIL import ImageFont
+from math import floor
+from time import sleep
+from approxeng.hwsupport import add_properties
 from luma.core.interface.serial import i2c
 from luma.core.render import canvas
 from luma.oled.device import ssd1306
 
-from redboard.magic import add_properties
-
 # Logger
 LOGGER = logging.getLogger(name='redboard')
-# logging.basicConfig(level=logging.DEBUG)
-
 
 # Potential I2C addresses for MX2 boards
 MX2_BOARD_CANDIDATE_I2C_ADDRESSES = [0x30, 0x31, 0x32, 0x33]
@@ -102,7 +100,7 @@ class RedBoardError(Exception):
 
 def i2c_device_exists(address, i2c_bus_number=1):
     try:
-        with smbus2.SMBus(bus=i2c_bus_number) as bus:
+        with SMBus(bus=i2c_bus_number) as bus:
             try:
                 bus.read_byte_data(i2c_addr=address, register=0)
                 return True
@@ -149,7 +147,7 @@ class MX2:
         :param speed:
             A value between -1.0 and 1.0, values outside this range will be clamped to it
         """
-        with smbus2.SMBus(self.i2c_bus_number) as bus:
+        with SMBus(self.i2c_bus_number) as bus:
             bus.write_i2c_block_data(i2c_addr=self.address,
                                      register=0x30 if motor == 0 else 0x40,
                                      data=[1 if speed >= 0 else 0, int(abs(speed) * 255)])
@@ -193,12 +191,11 @@ class RedBoard:
         """
 
         self._pi = None
-        self._pwm_frequency = pwm_frequency
         self._pwm_range = pwm_range
 
         # Configure PWM for the LED outputs
         for led_pin in [RedBoard.LED_R_PIN, RedBoard.LED_G_PIN, RedBoard.LED_B_PIN]:
-            self.pi.set_PWM_frequency(led_pin, self._pwm_frequency)
+            self.pi.set_PWM_frequency(led_pin, pwm_frequency)
             self.pi.set_PWM_range(led_pin, self._pwm_range)
 
         # Configure motor pulse and direction pins as outputs, set PWM frequency
@@ -207,9 +204,10 @@ class RedBoard:
             dir_pin = motor['dir']
             self.pi.set_mode(dir_pin, pigpio.OUTPUT)
             self.pi.set_mode(pwm_pin, pigpio.OUTPUT)
-            self.pi.set_PWM_frequency(pwm_pin, self._pwm_frequency)
+            self.pi.set_PWM_frequency(pwm_pin, pwm_frequency)
             self.pi.set_PWM_range(pwm_pin, self._pwm_range)
 
+        # Set bus number to use for ADC and MX2 boards
         self.i2c_bus_number = i2c_bus_number
 
         # Check for I2C based expansions. This is an array of I2C addresses for motor expansion boards. Each
@@ -225,13 +223,16 @@ class RedBoard:
         self.num_motors = len(RedBoard.MOTORS) + (len(self.i2c_motor_expansions) * 2)
 
         # Inject property based accessors, along with configuration infrastructure
-        add_properties(board=self, motors=range(0, self.num_motors), servos=RedBoard.SERVO_PINS, adcs=range(0, 4))
-
+        add_properties(board=self, motors=range(0, self.num_motors), servos=RedBoard.SERVO_PINS, adcs=[0, 1, 2, 3],
+                       default_adc_divisor=7891, leds=[0])
+        # Set adc0 divisor, this uses a different config as it's attached to the battery
         self.adc0_divisor = 1100
-        self.adc1_divisor = 7891
-        self.adc2_divisor = 7891
-        self.adc3_divisor = 7891
+        # Configure the LED, setting gamma and saturation correction to 2 for better pale colours, and brightness to 0.6
+        self.led0_brightness = 0.6
+        self.led0_gamma = 2
+        self.led0_saturation = 2
 
+        # Stop the motors if requested
         if stop_motors:
             self.stop()
         LOGGER.info('redboard initialised')
@@ -246,50 +247,30 @@ class RedBoard:
             self._pi = pigpio.pi()
         return self._pi
 
-    @staticmethod
-    def _check_positive(i):
-        f = float(i)
-        if f < 0.0:
-            LOGGER.warning('Value < 0, returning 0')
-            return 0.0
-        if f > 1.0:
-            LOGGER.warning('Value > 1.0, returning 1.0')
-            return 1.0
-        return f
+    def _set_led_rgb(self, led, r, g, b):
+        """
+        Set the on-board LED to the given r, g, b values - we only have one LED so we ignore the LED number. Values
+        range from 0.0 to 1.0, and have gamma, saturation, and brightness correction already applied.
+        """
+        self.pi.set_PWM_dutycycle(RedBoard.LED_R_PIN, r * self._pwm_range)
+        self.pi.set_PWM_dutycycle(RedBoard.LED_G_PIN, g * self._pwm_range)
+        self.pi.set_PWM_dutycycle(RedBoard.LED_B_PIN, b * self._pwm_range)
 
-    def set_led(self, h, s, v):
+    def _read_adc(self, adc, sleep_time=0.01):
         """
-        Set the on-board LED to the given hue, saturation, value (0.0-1.0)
-        """
-        r, g, b = colorsys.hsv_to_rgb(RedBoard._check_positive(h),
-                                      RedBoard._check_positive(s),
-                                      RedBoard._check_positive(v))
-        self.pi.set_PWM_dutycycle(RedBoard.LED_R_PIN, RedBoard._check_positive(r) * self._pwm_range)
-        self.pi.set_PWM_dutycycle(RedBoard.LED_G_PIN, RedBoard._check_positive(g) * self._pwm_range)
-        self.pi.set_PWM_dutycycle(RedBoard.LED_B_PIN, RedBoard._check_positive(b) * self._pwm_range)
-
-    def _read_adc(self, adc, divisor, digits=2, sleep_time=0.01):
-        """
-        Read from the onboard ADC. Note that ADC 0 is the battery monitor and will need a specific divisor to report
-        accurately, don't use this ADC with the default value for divisor unless you happen to have an exactly 3.3v
-        battery (kind of unlikely in this context, and it wouldn't work to power the RedBoard anyway)
+        Read a raw value from the onboard ADC
 
         :param adc:
             Integer index of the ADC to read, 0-3 inclusive
-        :param divisor:
-            Number to divide the reported value by to get a true voltage reading, defaults to 7891 for the 3.3v
-            reference
-        :param digits:
-            Number of digits to round the result, defaults to 2
         :param sleep_time:
             Time to sleep between setting the register from which to read and actually taking the reading. This is
             needed to allow the converter to settle, the default of 1/100s works for all cases, it may be possible
             to reduce it if needed for faster response, or consider putting this in a separate thread
         :return:
-            Measured voltage
+            Raw value from ADC channel
         """
         try:
-            with smbus2.SMBus(bus=self.i2c_bus_number) as bus:
+            with SMBus(bus=self.i2c_bus_number) as bus:
                 bus.write_i2c_block_data(RedBoard.ADC_I2C_ADDRESS, register=0x01,
                                          data=[RedBoard.ADC_REGISTER_ADDRESSES[adc], 0x83])
                 # ADC channels seem to need some time to settle, the default of 0.01 works for cases where we
@@ -297,12 +278,19 @@ class RedBoard:
                 if sleep_time:
                     time.sleep(sleep_time)
                 data = bus.read_i2c_block_data(RedBoard.ADC_I2C_ADDRESS, register=0x00, length=2)
+                return data[1] + (data[0] << 8)
         except FileNotFoundError as cause:
             raise RedBoardError('I2C not enabled, use raspi-config to enable and try again.') from cause
 
-        return round(float(data[1] + (data[0] << 8)) / divisor, ndigits=digits)
-
     def _set_servo_pulsewidth(self, servo_pin: int, pulse_width: int):
+        """
+        Set a servo pulse width
+
+        :param servo_pin:
+            Servo pin to set
+        :param pulse_width:
+            Pulse width in microseconds
+        """
         self.pi.set_servo_pulsewidth(servo_pin, pulse_width)
 
     def _set_motor_speed(self, motor, speed: float):
@@ -313,10 +301,8 @@ class RedBoard:
             The motor to set, 0 sets speed on motor A, 1 on motor B. If any I2C expansions are defined this will also
             accept higher number motors, where pairs of motors are allocated to each expansion address in turn.
         :param speed:
-            Speed between -1.0 and 1.0. If a value is supplied outside this range it will be clamped to this range
-            silently.
+            Speed between -1.0 and 1.0.
         """
-        LOGGER.debug(f'set motor{motor}={speed}')
         if motor < len(RedBoard.MOTORS):
             # Using the built-in motor drivers on the board
             self.pi.write(RedBoard.MOTORS[motor]['dir'], 1 if speed > 0 else 0)
@@ -326,7 +312,7 @@ class RedBoard:
             i2c_address = self.i2c_motor_expansions[(motor - len(RedBoard.MOTORS)) // 2]
             i2c_motor_number = (motor - len(RedBoard.MOTORS)) % 2
             try:
-                with smbus2.SMBus(bus=self.i2c_bus_number) as bus:
+                with SMBus(bus=self.i2c_bus_number) as bus:
                     bus.write_i2c_block_data(i2c_addr=i2c_address,
                                              register=0x30 if i2c_motor_number == 0 else 0x40,
                                              data=[1 if speed >= 0 else 0, int(abs(speed) * 255)])
@@ -335,13 +321,109 @@ class RedBoard:
 
     def _stop(self):
         """
-        SHUT IT DOWN! Equivalent to setting all motor speeds to zero and calling disable_servo on all available servo
-        outputs, then calling stop() on the PiGPIO instance. Any subsequent calls that would need pigpio will restart
-        it as a side effect, so it's safe to call this even if you've not completely finished with the board.
-
-        Also sets the speed on any connected I2C motor expansions to 0
+        Called by the injected stop() method after all motors, servos and LEDs have been deactivated, just cleans up
+        our PIGPIO instance.
         """
-        self.set_led(0, 0, 0)
-        self.pi.stop()
-        self._pi = None
-        LOGGER.info('RedBoard motors and servos stopped')
+        if self._pi is not None:
+            self._pi.stop()
+            self._pi = None
+        LOGGER.info('RedBoard motors, servos, and LED stopped')
+
+
+class PCA9685:
+    """
+    Drives boards based on the PCA9685 16 channel PWM chip. Code based on examples from Adafruit and elsewhere,
+    uses approxeng.hwsupport to manage configuration, property based access, range checking etc.
+    """
+
+    # When using its internal oscilator, this chip has a base clock of 25MHz, this is divided by
+    # a pre-scale value to derive the output PWM frequency. If an external oscillator is configured
+    # then that needs to be specified in the constructor, but this is the default value otherwise.
+    INTERNAL_CLOCK = 25000000.0
+
+    def __init__(self, i2c_bus_number=1, address=0x40, clock=INTERNAL_CLOCK):
+        """
+        Create a new driver for a PCA9685 chip on an accessible I2C bus
+
+        :param i2c_bus_number:
+            Bus number, defaults to 1 for the built-in hardware I2C on modern Raspberry Pi boards
+        :param address:
+            I2C address, set using solder jumpers on most expansion boards, defaults to 0x40 but use i2cdetect
+            to find the address on your particular system if needed
+        :param clock:
+            Oscillator for the chip, defaults to 25MHz for use with the internal clock source, you almost
+            certainly never need to set this explicitly
+        """
+        self._clock = clock
+        self._frequency = 0
+        self.i2c_bus_number = i2c_bus_number
+        self.address = address
+        self._mode = 0
+        self._frequency = 50
+        add_properties(board=self, servos=range(16))
+
+    @property
+    def frequency(self):
+        """
+        The PWM frequency used for channels on this board. Defaults to 50Hz which is a sensible value for most servos,
+        but you might want to change this if you're using the board to drive e.g. LEDs and need a faster update speed
+        to prevent flickering.
+        """
+        return self._frequency
+
+    @frequency.setter
+    def frequency(self, frequency_hz):
+        old_mode = self._mode
+        if old_mode is not None:
+            self._mode = (old_mode & 0x7F) | 0x10
+            # The prescale value must be at least 3, at most 255
+            prescale_value = min(255, max(3, int(floor(self._clock / (4096.0 * frequency_hz) - 0.5))))
+            with SMBus(bus=self.i2c_bus_number) as bus:
+                bus.write_byte_data(i2c_addr=self.address, register=0xFE,
+                                    value=prescale_value)
+            self._mode = old_mode
+            sleep(0.005)
+            self._mode = old_mode | 0x80
+            # Work out what the actual frequency will be now by inverting the prescale logic
+            # Most of the time this will be pretty close to the requested value, but we can
+            # make the mapping from duty cycle to pulse width more accurate if we store the
+            # actual rather than requested value here.
+            self._frequency = self._clock / (4096.0 * (prescale_value + 1))
+
+    @property
+    def _mode(self):
+        """
+        Command mode for the PCA9685, internal use.
+        """
+        with SMBus(bus=self.i2c_bus_number) as bus:
+            return bus.read_byte_data(i2c_addr=self.address, register=0x00)
+
+    @_mode.setter
+    def _mode(self, new_mode):
+        with SMBus(bus=self.i2c_bus_number) as bus:
+            bus.write_byte_data(i2c_addr=self.address, register=0x00, value=new_mode)
+
+    def _set_servo_pulsewidth(self, channel: int, pulse_width: int):
+        """
+        Pulse width specified in microseconds, channel should be 0..15. This is used by the hwsupport module
+        to create the various servoXXX properties and manage configuration. You probably don't want to use
+        it directly.
+        """
+        self.set_duty_cycle(channel, min(4095, int(pulse_width * self.frequency * (4096 / 1000000))))
+
+    def set_duty_cycle(self, channel: int, duty: int):
+        """
+        Directly set the duty cycle of a channel, use this if you're directly controlling the duty cycle as opposed
+        to the pulse width, for example to drive LEDs or other PWM peripherals where the duty cycle is more important
+        than the pulse width.
+
+        Note that there's no value checking on this method, you must set a channel between 0 and 15 inclusive and a
+        duty cycle from 0 to 4095 inclusive, values outside these ranges have undefined (and probably unfortunate)
+        results.
+
+        :param channel: channel 0..15
+        :param duty: duty cycle 0..4095
+        """
+        with SMBus(bus=self.i2c_bus_number) as bus:
+            bus.write_byte_data(i2c_addr=self.address, register=0x08 + channel * 4, value=duty & 0xFF)
+            bus.write_byte_data(i2c_addr=self.address, register=0x09 + channel * 4, value=duty >> 8)
